
.rst-content div[class^=highlight-good],
.rst-content div[class^=highlight-bad] {
    border-left: 2px solid transparent;
}

.rst-content div[class^=highlight-good] {
    border-left-color: rgba(74, 182, 93, 1);
}

.rst-content div[class^=highlight-bad]  {
    border-left-color: rgb(244, 76, 78);
}

/* side nav */
.wy-side-nav-search, .wy-nav-top {
    background: rgb(251, 251, 251);
    border-right: #e1e4e5 solid 1px;
}

/* version number text (side nav) */
.wy-side-nav-search>div.version {
    color: #2980b9;
}
/* size of logo in side nav */
.wy-side-nav-search > a img.logo, .wy-side-nav-search > a img.logo {
    width: 150px !important;
    height: auto !important;
}

/* TABLE SETTINGS */
/* override table no-wrap -> normal */
.rst-content table.docutils td, .rst-content table.docutils th {
    white-space: normal;
}
html.writer-html5 .rst-content table.docutils td>p, html.writer-html5 .rst-content table.docutils th>p {
    line-height: 24px;
}
/* suppress margins at bottom of cell contents */
.rst-content table.docutils ol:last-child {
    margin-bottom: 0;
}
.rst-content table.docutils div[class^=highlight] {
    margin-bottom: 0;
}

.rst-content .large-font-paragraph code  {
    font-size: 100%;
}